#include "FamilyLikelihoodLG.h"
#include <math.h>
#include <bitset>
#include <iostream>
#include <limits.h>

#define GLLIM 255
#define MIN_LLK -300
#define LK_ZERO 1e-35

using namespace std;

int FamilyLikelihoodLG::CompareDouble(const double * a, const double * b)
   {
   if (*a < *b) return -1;
   if (*a > *b) return 1;
   return 0;
   }

FamilyLikelihoodLG::FamilyLikelihoodLG()
{
  Initialize();
}

FamilyLikelihoodLG::~FamilyLikelihoodLG()
{
  Clear();
  if(marginals!=NULL)
    {
      for(int loc=0; loc<pm->count; loc++)
	delete [] marginals[loc];
      delete [] marginals;
      marginals = NULL;
    }
 for(int pidx1=0; pidx1<im->famSize; pidx1++)
    tmpMarkers[pidx1];
 delete [] tmpMarkers;
}

void FamilyLikelihoodLG::Initialize()
{
  pm = NULL;
  im = NULL;
  lvector = NULL;
  rvector = NULL;
  pvector = NULL;
  delta = NULL;
  phi = NULL;
  marginals = NULL;
  onfly=false;
  geno_labels.resize(3);
  geno_labels[0] = "0/0";
  geno_labels[1] = "0/1";
  geno_labels[2] = "1/1";

  phased_geno_labels.resize(4);
  phased_geno_labels[0] = "0|0";
  phased_geno_labels[1] = "0|1";
  phased_geno_labels[2] = "1|0";
  phased_geno_labels[3] = "1|1"; 
}

void FamilyLikelihoodLG::Clear()
{
  DeleteMemory();
}


void FamilyLikelihoodLG::Clear_Partial()
{
  if(lvector!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] lvector[i];
      delete [] lvector;
      lvector = NULL;
    }
  if(rvector!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] rvector[i];
      delete [] rvector;
      rvector = NULL;
    }
  
  if(pvector!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] pvector[i];
      delete [] pvector;
      pvector = NULL;
    }
  
  if(delta!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] delta[i];
      delete [] delta;
      delta = NULL;
    }
  
  if(phi!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] phi[i];
      delete [] phi;
      phi = NULL;
    }
 
}

void FamilyLikelihoodLG::DeleteMemory()
{
  if(lvector!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] lvector[i];
      delete [] lvector;
      lvector = NULL;
    }
  if(rvector!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] rvector[i];
      delete [] rvector;
      rvector = NULL;
    }
  
  if(pvector!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] pvector[i];
      delete [] pvector;
      pvector = NULL;
    }
  
  if(marginals!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] marginals[i];
      delete [] marginals;
      marginals = NULL;
    }
  
  if(delta!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] delta[i];
      delete [] delta;
      delta = NULL;
    }
  
  if(phi!=NULL)
    {
      for(int i=0; i<pm->count; i++)
	delete [] phi[i];
      delete [] phi;
      phi = NULL;
    }
}

void FamilyLikelihoodLG::SetPedMarkerData(PedMarkerData* m, int fid)
{
  pm = m;
  fam_idx = fid;
  //fam_markers = pm->markers[fam_idx];
  marker_count = pm->count;
  
}

void FamilyLikelihoodLG::SetInheritanceMatrix(Inheritance_Matrix* imp)
{
  im = imp;
  geno_prior.resize(im->four_pow_f);
  geno_posterior.Dimension(im->famSize, 3);
  best_geno_idx.resize(im->famSize);
  best_geno_label.resize(im->famSize); 
  geno_qual.resize(im->famSize);
  dosage.resize(im->famSize);

  best_phased_geno_idx.resize(im->famSize);
  best_phased_geno_label.resize(im->famSize); 
  phased_geno_qual = 0.0;

  geno_index_tmp.resize(im->four_pow_f);
  for(int i=0; i<geno_index_tmp.size(); i++)
   geno_index_tmp[i].resize(im->famSize);
}

void FamilyLikelihoodLG::AllocateMemory()
{
  if(im==NULL) {fprintf(stderr, "Inheritance vector pointer not set yet!\n"); exit(1); }
  if(pm==NULL) {fprintf(stderr, "Pedigree maker data pointer not set yet!\n"); exit(1); }

  DblTmpVector = new double[im->nInherit_vector];

  lvector = new float *[pm->count];
  for(int i=0; i<pm->count; i++)
    lvector[i] = new float[im->nInherit_vector];

  rvector = new float *[pm->count];
  for(int i=0; i<pm->count; i++)
    rvector[i] = new float[im->nInherit_vector];

  pvector = new float *[pm->count];
  for(int i=0; i<pm->count; i++)
    pvector[i] = new float[im->nInherit_vector];

 tmpMarkers = new SeqMarkerInfo*[im->famSize];
 for(int pidx1=0; pidx1<im->famSize; pidx1++)
    tmpMarkers[pidx1] = new SeqMarkerInfo[im->famSize];

  pow2.resize(im->two_n_minus_f);
  for(int i=0; i<pow2.size(); i++)
    pow2[i] = pow(2, i);

  //lk_single_point.resize(pm->count); // not used now
 llk_M1_partition.resize(im->famSize);
}

float FamilyLikelihoodLG::GetSinglePointLikelihood(int loc)
{
  return(lk_single_point[loc]); // lk_single_point is not calculated for now
}

// calculate ped lk assuming iv uniform and is equivalanet to peeling algo
// not used for now
float FamilyLikelihoodLG::CalculateSinglePointLikelihood(int loc)
{
  float lk = 0.0;
  SetGenoPrior(pm->var[loc].freq);

 std::vector<std::vector<uint8_t> > geno_index_onfly;
 geno_index_onfly = geno_index_tmp;

  #pragma omp parallel for reduction(+:lk) firstprivate(geno_index_onfly)
  for(int iv=0; iv<im->nInherit_vector; iv++)
   {
    if(onfly) lk += CalculateSingleVectorLikelihood(iv, geno_index_onfly, loc);
    else lk += CalculateSingleVectorLikelihood(iv, loc);
   }
 return(lk);
}

float FamilyLikelihoodLG::CalculateRefLogLikelihood(SeqMarkerInfo *markers)
{
 float llk = 0;
 for(int i=0; i<im->famSize; i++)
  llk += (-0.1*markers[i].geno_pl[0]);

 return(llk);
}

// max_iv_idx is pre-computed and is used for nearby variants
void FamilyLikelihoodLG::CalculatePhasedGeno_best(SeqMarkerInfo *markers, int max_iv_idx)
{
 int max_hap_idx = CalculateSingleVectorMaxHapLikelihood(max_iv_idx, markers);

 int a1; int a2;
 ac = 0; dp = 0;

 for(int pidx=0; pidx<im->famSize; pidx++)
 {
   a1 = im->founder_node_indicator[max_hap_idx][im->inherit_matrix[max_iv_idx][pidx*2]];
   a2 = im->founder_node_indicator[max_hap_idx][im->inherit_matrix[max_iv_idx][pidx*2+1]];

   best_phased_geno_idx[pidx] = a1*2+a2;
   best_phased_geno_label[pidx] = phased_geno_labels[best_phased_geno_idx[pidx]];

   ac += (a1 + a2);
   dp += markers[pidx].missing_dp ? 0 : markers[pidx].dp; 
 }
}

// No need to find max marignal for each variant. Use the pre-computed best indices above. This function is not used here
void FamilyLikelihoodLG::CalculatePhasedGeno_marginal(SeqMarkerInfo *markers, float *marginals)
{
 int max_iv_idx = 0;
 float max_llk = 0;
 for(int i=0; i<im->nInherit_vector; i++) {
  if(max_llk<marginals[i]) { max_llk = marginals[i]; max_iv_idx = i; }
 }

 int max_hap_idx = CalculateSingleVectorMaxHapLikelihood(max_iv_idx, markers);

 int a1; int a2;
 ac = 0; dp = 0;

 for(int pidx=0; pidx<im->famSize; pidx++)
 {
   a1 = im->founder_node_indicator[max_hap_idx][im->inherit_matrix[max_iv_idx][pidx*2]];
   a2 = im->founder_node_indicator[max_hap_idx][im->inherit_matrix[max_iv_idx][pidx*2+1]];

   best_phased_geno_idx[pidx] = a1*2+a2;
   best_phased_geno_label[pidx] = phased_geno_labels[best_phased_geno_idx[pidx]];

   ac += (a1 + a2);
   dp += markers[pidx].missing_dp ? 0 : markers[pidx].dp; 
 }
}

void FamilyLikelihoodLG::CalculateGenoPosteriors(SeqMarkerInfo *markers, float *marginals, int max_idx, bool single_marginal)
{
  for(int pidx1=0; pidx1<im->famSize; pidx1++)
   for(int pidx2=0; pidx2<im->famSize; pidx2++)
    tmpMarkers[pidx1][pidx2] = markers[pidx2];

  std::vector<int> ac_ind;
  ac_ind.resize(im->famSize);

  float geno_lk[3];

  #pragma omp parallel for firstprivate(geno_lk)
  for(int pidx=0; pidx<im->famSize; pidx++)
    {

      geno_lk[0] = CalculateSinglePointLikelihood_IBD_Geno(tmpMarkers[pidx], marginals, fam_idx, pidx, 0, max_idx, single_marginal);

      geno_lk[1] = CalculateSinglePointLikelihood_IBD_Geno(tmpMarkers[pidx], marginals, fam_idx, pidx, 1, max_idx, single_marginal);

      geno_lk[2] = CalculateSinglePointLikelihood_IBD_Geno(tmpMarkers[pidx], marginals, fam_idx, pidx, 2, max_idx, single_marginal);

      float sum = geno_lk[0] + geno_lk[1] + geno_lk[2];

      float max_lk = 0.0;
      int best = 0;

      for(int i=0; i<3; i++) if(max_lk<geno_lk[i]) { max_lk = geno_lk[i]; best = i; }

      if(sum>0) {
      geno_posterior[pidx][0] = geno_lk[0]/sum;
      geno_posterior[pidx][1] = geno_lk[1]/sum;
      geno_posterior[pidx][2] = geno_lk[2]/sum;
      best_geno_idx[pidx] = best;
      best_geno_label[pidx] = geno_labels[best];
      geno_qual[pidx] = geno_posterior[pidx][best] > 0.9999999999 ? 100 : -10*log10(1-geno_posterior[pidx][best]);
      dosage[pidx] = geno_posterior[pidx][1] + 2*geno_posterior[pidx][2];

     if(best==1) ac_ind[pidx]++;
     if(best==2) ac_ind[pidx]+=2;
     }
     else {
      geno_posterior[pidx][0] = geno_posterior[pidx][1] = geno_posterior[pidx][2] = 0;
      best_geno_idx[pidx] = 0;
      best_geno_label[pidx] = "./.";
      geno_qual[pidx] = 0;
      dosage[pidx] = 0;
    }
  } //end of for loop for all individuals in a family
  ac=0; dp=0; for(int i=0; i<ac_ind.size(); i++) { ac += ac_ind[i]; dp += markers[i].missing_dp ? 0 : markers[i].dp; }
}

// De novo mutation using posterior IBD sharing at the variant site on the map
float FamilyLikelihoodLG::CalculateSinglePointLikelihood_IBD_M1(int loc)
{
  float lk = 0.0;

  #pragma omp parallel for reduction(+:lk)
  for(int iv=0; iv<im->nInherit_vector; iv++)
  {
    lk += marginals[loc][iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood_denovo(iv, loc) * marginals[loc][iv];
  }
  return(lk);
}

float FamilyLikelihoodLG::CalculateSinglePointLikelihood_IBD(int loc)
{
  float lk = 0.0;

  //SetGenoPrior(pm->var[loc].freq);
 std::vector<std::vector<uint8_t> > geno_index_onfly;
 geno_index_onfly = geno_index_tmp;

  #pragma omp parallel for reduction(+:lk) firstprivate(geno_index_onfly)
  for(int iv=0; iv<im->nInherit_vector; iv++)
  {
    if(onfly) lk += marginals[loc][iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood(iv, geno_index_onfly, loc) * marginals[loc][iv];
    else lk += marginals[loc][iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood(iv, loc) * marginals[loc][iv];
  }
  return(lk);
}

// Single point likelhood of pedigree using a single posterior iv to speedup for best_path
float FamilyLikelihoodLG::CalculateSinglePointLikelihood_IBD(int loc, int iv)
{
  float lk = 0.0;

  //SetGenoPrior(pm->var[loc].freq);
 std::vector<std::vector<uint8_t> > geno_index_onfly;
 geno_index_onfly = geno_index_tmp;

    if(onfly) lk += marginals[loc][iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood(iv, geno_index_onfly, loc) * marginals[loc][iv];
    else lk += marginals[loc][iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood(iv, loc) * marginals[loc][iv];

  return(lk);
}

// De novo mutation
float FamilyLikelihoodLG::CalculateSinglePointLikelihood_IBD_M1(SeqMarkerInfo *mk, float *marginals)
{
  float lk = 0.0;
  //SetGenoPrior(freq);

  #pragma omp parallel for reduction(+:lk)
  for(int iv=0; iv<im->nInherit_vector; iv++)
  {
   lk += marginals[iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood_denovo(iv, mk) * marginals[iv];
  }
  return(lk);

}

// Single point likelhood of pedigree using posterior IBD sharing
float FamilyLikelihoodLG::CalculateSinglePointLikelihood_IBD(SeqMarkerInfo *mk, float *marginals)
{
  float lk = 0.0;
  //SetGenoPrior(freq);

 std::vector<std::vector<uint8_t> > geno_index_onfly;
 geno_index_onfly = geno_index_tmp;

  #pragma omp parallel for reduction(+:lk) firstprivate(geno_index_onfly)
  for(int iv=0; iv<im->nInherit_vector; iv++)
  {
    if(onfly) lk += marginals[iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood(iv, geno_index_onfly, mk) * marginals[iv];
    else lk += marginals[iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood(iv, mk) * marginals[iv];
  }
  return(lk);
}

// Single point likelhood of pedigree using a single posterior IBD to speed up calculation using best_path
float FamilyLikelihoodLG::CalculateSinglePointLikelihood_IBD(SeqMarkerInfo *mk, float *marginals, int iv)
{
  float lk = 0.0;
  //SetGenoPrior(freq);

 std::vector<std::vector<uint8_t> > geno_index_onfly;
 geno_index_onfly = geno_index_tmp;

 if(onfly) lk += marginals[iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood(iv, geno_index_onfly, mk) * marginals[iv];
 else lk += marginals[iv]==0.0 ? 0.0 : CalculateSingleVectorLikelihood(iv, mk) * marginals[iv];

 return(lk);
}

// Single point likelhood of an individual (fidx and pidx) for a genotype (gidx) with posterior IBD sharing. Here max_idx is the loc idx with max lk when a variant is in between
float FamilyLikelihoodLG::CalculateSinglePointLikelihood_IBD_Geno(SeqMarkerInfo *markers, float *marginals, int fidx, int pidx, int gidx, int max_idx, bool single_marginal)
{
  float lk = 0.0;
  //SetGenoPrior(freq);

  int zero_idx1;
  int zero_idx2;
  float geno_lk1, geno_lk2;
  uint8_t geno_pl1, geno_pl2;

  switch(gidx)
    {
    case 0:
      zero_idx1 = 1;
      zero_idx2 = 2;
      break;
    case 1:
      zero_idx1 = 0;
      zero_idx2 = 2;
      break;
    case 2:
      zero_idx1 = 0;
      zero_idx2 = 1;
      break;
    default:
      error("GIDX has to be 0, 1 or 2");
    }

  geno_lk1 = markers[pidx].geno_lk[zero_idx1];
  geno_lk2 = markers[pidx].geno_lk[zero_idx2];
  geno_pl1 = markers[pidx].geno_pl[zero_idx1];
  geno_pl2 = markers[pidx].geno_pl[zero_idx2];

  markers[pidx].geno_lk[zero_idx1] = pmd.PL2LK(GLLIM);
  markers[pidx].geno_lk[zero_idx2] = pmd.PL2LK(GLLIM);
  markers[pidx].geno_pl[zero_idx1] = GLLIM;
  markers[pidx].geno_pl[zero_idx2] = GLLIM;

 std::vector<std::vector<uint8_t> > geno_index_onfly;
 geno_index_onfly = geno_index_tmp;

  if(single_marginal)
   lk = CalculateSingleVectorLikelihood(best_marginal[max_idx], markers);
  else {
  #pragma omp parallel for reduction(+:lk) firstprivate(geno_index_onfly)
  for(int iv=0; iv<im->nInherit_vector; iv++)
   {
    if(onfly) lk += marginals[iv]==0.0 ? 0.0 : ( CalculateSingleVectorLikelihood(iv, geno_index_onfly, markers) * marginals[iv]);
    else lk += marginals[iv]==0.0 ? 0.0 : ( CalculateSingleVectorLikelihood(iv, markers) * marginals[iv]);
   }
  }

  markers[pidx].geno_lk[zero_idx1] = geno_lk1;
  markers[pidx].geno_lk[zero_idx2] = geno_lk2;
  markers[pidx].geno_pl[zero_idx1] = geno_pl1;
  markers[pidx].geno_pl[zero_idx2] = geno_pl2;

  return(lk);
}

//Priors are genotype frequecies assusing HWE
void FamilyLikelihoodLG::SetGenoPrior(float &freq)
{
  int bc;
  for(int i=0; i<im->four_pow_f; i++)
    {
      bc = im->Count_Bits(i);
      geno_prior[i] = pow(freq, im->two_f-bc) * pow(1-freq, bc);
    }
}

//Sum over inheritance vectors for a single founder alleles configuration
float FamilyLikelihoodLG::CalculateSingleFounderAllelesLikelihood(int founder_alleles_idx, int loc)
{
  double lk = .0;
  float llk = .0;
  
  for(int i=0; i<im->nInherit_vector; i++)
    {
      llk = .0;
      for(int j=0; j<im->famSize; j++)
	llk += (-0.1*pm->markers[loc][fam_idx][j].geno_pl[im->geno_index[i][founder_alleles_idx][j]]); 
      
      lk += ( pow(10, llk) * geno_prior[i] );
    }
  return(lk);
}

// Only ref is allowed for founders for a single inheritance vector
float FamilyLikelihoodLG::CalculateSingleVectorLikelihood_denovo(int iv_index, int loc)
{
  double lk = .0;
  float llk = .0;

   for(int i=2; i<im->famSize; i++)
  {
     llk_M1_partition[i-2] = 0.0;
     for(int j=0; j<im->famSize; j++)
      {
        if(j==i)
          llk_M1_partition[i-2] += (-0.1*pm->markers[loc][fam_idx][j].geno_pl[1]);

        else
         llk_M1_partition[i-2] += (-0.1*pm->markers[loc][fam_idx][j].geno_pl[im->geno_index[iv_index][0][j]]);
      }

      llk_M1_partition[i-2] += log10(geno_prior[0]);
      lk += pow(10, llk_M1_partition[i-2]);
   }
  return(lk/(im->famSize-2));
}

//Sum over founder allele configurations for a single inheritance vector
float FamilyLikelihoodLG::CalculateSingleVectorLikelihood(int iv_index, int loc)
{
  double lk = .0;
  float llk = .0;
  
  // This function is called in a for loop and outer loop has been parallelized. So no parallelization here  
  for(int i=0; i<im->four_pow_f; i++)
    {
      llk = .0;
      for(int j=0; j<im->famSize; j++)
	{
	  llk += (-0.1*pm->markers[loc][fam_idx][j].geno_pl[im->geno_index[iv_index][i][j]]); 
	}
      lk += ( pow(10, llk) * geno_prior[i] );
    }
  return(lk);
}

//Sum over founder allele configurations for a single inheritance vector, using geno_index_onfly to save memory
float FamilyLikelihoodLG::CalculateSingleVectorLikelihood(int iv_index, std::vector<std::vector<uint8_t> >& geno_index_onfly, int loc)
{
  double lk = .0;
  float llk = .0;

  if(onfly) im->CalcFounderGenoMappingByIV(iv_index, geno_index_onfly);

  // This function is called in a for loop and outer loop has been parallelized. So no parallelization here  
  for(int i=0; i<im->four_pow_f; i++)
    {
      llk = .0;
      for(int j=0; j<im->famSize; j++)
	{
	  if(onfly) llk += (-0.1*pm->markers[loc][fam_idx][j].geno_pl[geno_index_onfly[i][j]]); 
	  else llk += (-0.1*pm->markers[loc][fam_idx][j].geno_pl[im->geno_index[iv_index][i][j]]); 
	}
      lk += ( pow(10, llk) * geno_prior[i] );

    }
  return(lk);
}

//Only ref is considered in founders for a single inheritance vector
float FamilyLikelihoodLG::CalculateSingleVectorLikelihood_denovo(int iv_index, SeqMarkerInfo *markers)
{
  double lk = .0;
  float llk = .0;

   for(int i=2; i<im->famSize; i++)
  {
     llk_M1_partition[i-2] = 0.0;
     for(int j=0; j<im->famSize; j++)
      {
        if(j==i)
	  llk_M1_partition[i-2] += (-0.1*markers[j].geno_pl[1]);

	else
	 llk_M1_partition[i-2] += (-0.1*markers[j].geno_pl[im->geno_index[iv_index][0][j]]);
      }

      llk_M1_partition[i-2] += log10(geno_prior[0]);
      lk += pow(10, llk_M1_partition[i-2]);
   }

  return(lk/(im->famSize-2));
}

//Sum over founder allele configurations for a single inheritance vector
float FamilyLikelihoodLG::CalculateSingleVectorLikelihood(int iv_index, SeqMarkerInfo *markers)
{
  double lk = .0;
  float llk = .0;

  // This function is called in a for loop and outer loop has been parallelized. So no parallelization here
  for(int i=0; i<im->four_pow_f; i++)
    {
      llk = .0;
      for(int j=0; j<im->famSize; j++)
	{
	  llk += (-0.1*markers[j].geno_pl[im->geno_index[iv_index][i][j]]);
	}

      lk += ( pow(10, llk)* geno_prior[i] );
    }

  return(lk);
}

//Sum over founder allele configurations for a single inheritance vector, using geno_index_onfly to same memory
float FamilyLikelihoodLG::CalculateSingleVectorLikelihood(int iv_index, std::vector<std::vector<uint8_t> >& geno_index_onfly, SeqMarkerInfo *markers)
{
  double lk = .0;
  float llk = .0;

  if(onfly) im->CalcFounderGenoMappingByIV(iv_index, geno_index_onfly);

  // This function is called in a for loop and outer loop has been parallelized. So no parallelization here
  for(int i=0; i<im->four_pow_f; i++)
    {
      llk = .0;
      for(int j=0; j<im->famSize; j++)
	{
	  if(onfly) llk += (-0.1*markers[j].geno_pl[geno_index_onfly[i][j]]);
	  else llk += (-0.1*markers[j].geno_pl[im->geno_index[iv_index][i][j]]);
	}
      lk += (pow(10, llk) * geno_prior[i] );
    }

  return(float(lk));
}

//Find the founder allele configuration with the MAX likelihood for a single inheritance vector
int FamilyLikelihoodLG::CalculateSingleVectorMaxHapLikelihood(int iv_index, SeqMarkerInfo *markers)
{
  float max_lk = 0.0;
  int max_lk_idx = 0;
  float lk = .0;
  float total_lk = .0;

  // This function is called in a for loop and outer loop has been parallelized. So no parallelization here
  for(int i=0; i<im->four_pow_f; i++)
    {
      float llk = .0;
      for(int j=0; j<im->famSize; j++)
	{
	  llk += (-0.1*markers[j].geno_pl[im->geno_index[iv_index][i][j]]);
	}
      lk = pow(10, llk) * geno_prior[i];
      total_lk += lk;

      if(i==0 || max_lk<lk) { max_lk=lk; max_lk_idx=i; }
    }
  phased_geno_qual = total_lk<=0.0 ? 0 : (max_lk/total_lk>0.9999999999 ? 100 : -10*log10(1-max_lk/total_lk));

  return(max_lk_idx);
}

void FamilyLikelihoodLG::CalculateIVProb()
{
  int cnt=0;

  for(int loc=0; loc<pm->count; loc++)
    {
      SetGenoPrior(pm->var[loc].freq);

      std::vector<std::vector<uint8_t> > geno_index_onfly;
      geno_index_onfly = geno_index_tmp;

      double sum = 0.0;
      #pragma omp parallel for reduction(+:sum) firstprivate(geno_index_onfly)
      for(int iv=0; iv<im->nInherit_vector; iv++)
	{
	  if(onfly) pvector[loc][iv] = CalculateSingleVectorLikelihood(iv, geno_index_onfly, loc);
	  else pvector[loc][iv] = CalculateSingleVectorLikelihood(iv, loc);

	  sum+= pvector[loc][iv];
	  //lk_single_point[loc] += pvector[loc][iv]; // not used for now
	}

    if(sum<LK_ZERO) for(int iv=0; iv<im->nInherit_vector; iv++) pvector[loc][iv] = 1.0;

    if(++cnt%1000==0) printf("Finishing %d markers in calculating inheritance vector probabilities...\n", cnt);
    }
}

// Naive implementation and replaced by FFT
void FamilyLikelihoodLG::CalculateMultiPointLikelihood_forward()
{
  int bc = 0;
  float lk = 0.0;

  if(im->bit_count_trans.size()==0) im->Count_Trans_Bits();

  if(scaling_factors.size()==0) scaling_factors.resize(pm->count);

  for(int loc=0; loc<pm->count; loc++)
    {
      if(loc==0)
	{
	  for(int iv=0; iv<im->nInherit_vector; iv++)   
	    lvector[loc][iv] = pvector[loc][iv]/im->nInherit_vector; 
	  
	  scaling_factors[loc] = Scale(lvector[loc]);
	  
	  continue;
	}
     
      #pragma omp parallel for firstprivate(bc, lk)
      for(int i=0; i<im->nInherit_vector; i++)
	{ 
	  lk = 0.;
          //if(pm->marker_map.recomb[loc-1]==0.0) lk = lvector[loc-1][i];
	  //else 
           for(int j=0; j<im->nInherit_vector; j++)
	    {
	      bc = im->bit_count_trans[im->inherit_vector[i]^im->inherit_vector[j]];
	      lk += (lvector[loc-1][j] * pow(pm->marker_map.recomb[loc-1], bc) * pow(1-pm->marker_map.recomb[loc-1], im->two_n_minus_f-bc) );
	    }

	  lvector[loc][i] = lk * pvector[loc][i];
	}
      scaling_factors[loc] = Scale(lvector[loc]);

     if(loc>0 && loc%10000==0) printf("Finishing %d marker in forward probabilities...\n", loc+1);
    }
}

// Naive implementation and replaced by FFT
void FamilyLikelihoodLG::CalculateMultiPointLikelihood_backward()
{
  int bc;
  float lk = 0.0;

  if(im->bit_count_trans.size()==0) im->Count_Trans_Bits();

  for(int loc=pm->count-1; loc>=0; loc--)
    {
      if(loc==pm->count-1)
	{
	  for(int i=0; i<im->nInherit_vector; i++)
	    rvector[loc][i] = 1.;
	    Scale(rvector[loc], scaling_factors[loc]);
	  continue;
	}
      
      //for the second to the last and backword
      #pragma omp parallel for firstprivate(bc, lk)

      for(int i=0; i<im->nInherit_vector; i++)
	{
	  lk = 0.0;
          //if(pm->marker_map.recomb[loc]==0.0) lk = rvector[loc+1][i];
	  //else
	   for(int j=0; j<im->nInherit_vector; j++)
	    {
	      bc = im->bit_count_trans[im->inherit_vector[i]^im->inherit_vector[j]];
	      lk += (rvector[loc+1][j] * pow(pm->marker_map.recomb[loc], bc) * pow(1-pm->marker_map.recomb[loc], im->two_n_minus_f-bc) * pvector[loc+1][j] );
	    }
	  rvector[loc][i] = lk;
	} // end of outer loop for iv;

      Scale(rvector[loc], scaling_factors[loc]);

     if(loc>0 && loc%10000==0) printf("Finishing %d marker in backward probabilities...\n", loc+1);


    }// end of loc for loop
  
}

void FamilyLikelihoodLG::CalculateMarginals()
{
  if(tmpMarginals.size()==0) tmpMarginals.resize(im->nInherit_vector);

  if(marginals==NULL)
    {
      marginals = new float*[pm->count];
      for(int i=0; i<pm->count; i++)
	marginals[i] = new float[im->nInherit_vector];
    }

  for(int loc=0; loc<pm->count; loc++) {
    for(int i=0; i<im->nInherit_vector; i++) {
       marginals[loc][i] = lvector[loc][i] * rvector[loc][i];
//printf("lv=%e rv=%e marg=%e\n", lvector[loc][i], rvector[loc][i], marginals[loc][i]);
if(lvector[loc][i]<0.0) error("lvector[%d][%d]=%e\n", loc, i, lvector[loc][i]);
//if(rvector[loc][i]<0.0) error("rvector[%d][%d]=%e\n", loc, i, rvector[loc][i]);

}
    Scale(marginals[loc]);
    if(cumulative_marginals<1.0) MaskMarginals(loc);
  }
}

void FamilyLikelihoodLG::FindBestMarginal()
{
  if(best_marginal.size()==0) best_marginal.resize(pm->count);

  float max_marg;
  int best;
  for(int loc=0; loc<pm->count; loc++)
  {
    max_marg = marginals[loc][0];
    best = 0;

    for(int i=0; i<im->nInherit_vector; i++)
    {
     if(max_marg<marginals[loc][i])
      {
        max_marg = marginals[loc][i];
        best = i;
      }
    }
   best_marginal[loc] = best;
  }
}

void FamilyLikelihoodLG::FindBestPath()
{

  if(im->bit_count_trans.size()==0) im->Count_Trans_Bits();

  if(delta==NULL)
    {
      delta = new float *[pm->count];
      for(int i=0; i<pm->count; i++)
	delta[i] = new float[im->nInherit_vector];
    }
  
  phi = new uint32_t *[pm->count];
  for(int i=0; i<pm->count; i++)
    phi[i] = new uint32_t[im->nInherit_vector];
  
  if(best_path.size()==0) best_path.resize(pm->count);
  
  float lk, max_lk;
  int max_iv;
  int bc;
  
  for(int loc=0; loc<pm->count; loc++)
    {
      if(loc==0)
	{
	  for(int i=0; i<im->nInherit_vector; i++)
	    {
	      delta[loc][i] = log10(pvector[loc][i]);
	      phi[loc][i] = 0;
	    }
	  continue;
	}

      #pragma omp parallel for firstprivate(bc, lk, max_lk, max_iv)
      for(int i=0; i<im->nInherit_vector; i++)
	{
         max_lk = .0;
         max_iv = 0;
         lk = .0;
         if(pm->marker_map.recomb[loc-1]==0.0)
           {
             max_lk = delta[loc-1][i];
             max_iv = i;
           }
         else
	  for(int j=0; j<im->nInherit_vector; j++)
	    {
	      bc = im->bit_count_trans[im->inherit_vector[i] ^ im->inherit_vector[j]];
	      lk = delta[loc-1][j] + bc*log10(pm->marker_map.recomb[loc-1]) + (im->two_n_minus_f-bc)*log10(1-pm->marker_map.recomb[loc-1]);
	      if(max_lk<lk || j==0) { max_lk = lk; max_iv = j; }
	    }
	  delta[loc][i] = max_lk + log10(pvector[loc][i]);
	  phi[loc][i] = max_iv;
	}

	 if(loc>0 && loc%10000==0) printf("Finishing %d markers in finding the best path...\n", loc+1);

    }
  
  best_path[pm->count-1] = 0;
  max_lk = delta[pm->count-1][0];

  for(int i=0; i<im->nInherit_vector; i++)
    if(max_lk<delta[pm->count-1][i])
      {
	best_path[pm->count-1] = i;
	max_lk = delta[pm->count-1][i];
      }

  for(int loc=pm->count-2; loc>=0; loc--) {
    best_path[loc] = phi[loc+1][best_path[loc+1]];
  }
}

void FamilyLikelihoodLG::PrintForwardProb()
{
  for(int loc=0; loc<pm->count; loc++)
    {
      for(int i=0; i<im->nInherit_vector; i++)
	printf("%f ", lvector[loc][i]);
      printf("\n");
    }
 fflush(stdout);
}

void FamilyLikelihoodLG::PrintBackwardProb()
{
  for(int loc=0; loc<pm->count; loc++)
    {
      for(int i=0; i<im->nInherit_vector; i++)
	printf("%f ", rvector[loc][i]);
      printf("\n");
    }
 fflush(stdout);

}

void FamilyLikelihoodLG::PrintMarginalProb()
{
  for(int loc=0; loc<pm->count; loc++)
    {
      for(int i=0; i<im->nInherit_vector; i++)
	printf("%f ", marginals[loc][i]);
      printf("\n");
    }
 fflush(stdout);
}

void FamilyLikelihoodLG::PrintBestPath()
{
  for(int loc=0; loc<pm->count; loc++)
    {
      printf("%d:%d:", loc+1, best_path[loc]); 
      cout<<bitset<10>(im->inherit_vector[best_path[loc]]).to_string()<<endl; //
    }
  printf("\n");
 fflush(stdout);
}

void FamilyLikelihoodLG::CalculateMultiPointLikelihood_forward_FFT()
{
  error("This function should not be called...\n");

 if(scaling_factors.size()==0) scaling_factors.resize(pm->count);

  float *tmpvector;
  tmpvector = new float[im->nInherit_vector];
  for(int iv=0; iv<im->nInherit_vector; iv++) lvector[0][iv] = pvector[0][iv]/im->nInherit_vector;
  scaling_factors[0] = Scale(lvector[0]);

  for(int loc=0; loc<pm->count-1; loc++)
    {
/*
      if(loc==0)
	{
	  for(int iv=0; iv<im->nInherit_vector; iv++){
	    lvector[loc][iv] = pvector[loc][iv]/im->nInherit_vector;
	    tmpvector[iv] = lvector[loc][iv];
          }
          scaling_factors[loc] = Scale(lvector[loc]);
	  continue;
	}
*/
	for(int iv=0; iv<im->nInherit_vector; iv++)
	  tmpvector[iv] = lvector[loc][iv];

      Convolve_FFT(pm->marker_map.recomb[loc], tmpvector);

      for(int iv=0; iv<im->nInherit_vector; iv++)
	lvector[loc+1][iv] = tmpvector[iv]*pvector[loc+1][iv];

      scaling_factors[loc+1] = Scale(lvector[loc+1]);
    }
  delete tmpvector;
}

void FamilyLikelihoodLG::CalculateMultiPointLikelihood_backward_FFT()
{
 if(scaling_factors.size()==0) scaling_factors.resize(pm->count);

  float *tmpvector;
  tmpvector = new float[im->nInherit_vector];
  for(int iv=0; iv<im->nInherit_vector; iv++) rvector[pm->count-1][iv] = pvector[pm->count-1][iv];
  Scale(rvector[pm->count-1], scaling_factors[pm->count-1]);

  for(int loc=pm->count-1; loc>0; loc--)
    {
/*      if(loc==pm->count-1)
	{
	  for(int iv=0; iv<im->nInherit_vector; iv++) {
	    rvector[loc][iv] = pvector[loc][iv];
            tmpvector[iv] = pvector[loc][iv];
          }

          Scale(rvector[loc], scaling_factors[loc]);
	  continue;
	}
*/

      for(int iv=0; iv<im->nInherit_vector; iv++) 
         tmpvector[iv] = rvector[loc][iv];

      Convolve_FFT(pm->marker_map.recomb[loc-1], tmpvector);

      for(int iv=0; iv<im->nInherit_vector; iv++)
	rvector[loc-1][iv] = tmpvector[iv]*pvector[loc-1][iv];

      Scale(rvector[loc-1], scaling_factors[loc-1]);

    }
  delete tmpvector;
}

float FamilyLikelihoodLG::Scale(float *pv)
{
  double sum=0.;
  for(int iv=0; iv<im->nInherit_vector; iv++)
    sum += pv[iv];
  for(int iv=0; iv<im->nInherit_vector; iv++)
    pv[iv] /= sum;

  return(sum);
}

float FamilyLikelihoodLG::Scale(double *source, float *pv)
{
  double sum=0.;
  for(int iv=0; iv<im->nInherit_vector; iv++)
    sum += source[iv];
  for(int iv=0; iv<im->nInherit_vector; iv++) {
    pv[iv] = source[iv]/sum;
    printf("Scale: pv[%d]=%e source=%e sum=%e\n", iv, pv[iv], source[iv], sum); }
  return(sum);
}

void FamilyLikelihoodLG::MaskMarginals(int loc)
{
   for(int i=0; i<im->nInherit_vector; i++) { tmpMarginals[i] = marginals[loc][i]; }

   std::sort (tmpMarginals.begin(), tmpMarginals.end());

   double sum = 0;
   float cutoff = 0.0; //= tmpMarginals[0];
   for(int i=0; i<im->nInherit_vector; i++)
   {
     sum+=tmpMarginals[i];
     if(sum >= 1-cumulative_marginals)
     {
	if(i==0) cutoff = 0.0;
	else cutoff = tmpMarginals[i-1];
	break;
     }
   }

  for(int i=0; i<im->nInherit_vector; i++)
    if(marginals[loc][i]<=cutoff) marginals[loc][i] = 0.0;
}

void FamilyLikelihoodLG::Scale(float *pv, float factor)
{
  for(int iv=0; iv<im->nInherit_vector; iv++)
    pv[iv] /= factor;
}

void FamilyLikelihoodLG::Convolve_FFT(float theta, float *pv)
{
  unsigned int i,j;

  if(trans.size()==0) trans.resize(im->nInherit_vector);

/*
  float one_minus_two_theta = 1.0 - 2.0*theta;

  for(i=0; i<im->nInherit_vector; i++)
   trans[i] = pow(one_minus_two_theta, im->bit_count[i]);
*/
    trans[0] = 1.0;
    double factor = 1.0 - 2.0*theta;
    for (j=1; j<=im->two_n; ++j)
      trans[j] = trans[j-1]*factor;

  FFT(pv);

  for (i=0; i<im->nInherit_vector; i++)
    pv[i] *= trans[im->bit_count[i]];

  FFT(pv);

  for(i=0; i<im->nInherit_vector; i++)
    pv[i] /= im->nInherit_vector;
}

void FamilyLikelihoodLG::FFT(float *pv)
{
        int i,j,ej,flipi;
        double temp;

        for (j=0; j<im->two_n_minus_f; j++) {
                ej = pow2[j];
                for (i=0; i<im->nInherit_vector; i++) {
                        if ((i & ej) == 0) {
                                flipi = i ^ ej;
                                temp = pv[i];
                                pv[i] = temp + pv[flipi];
                                pv[flipi] = temp - pv[flipi];
                        }
                }
        }
}

void FamilyLikelihoodLG::quickSort(float arr[], int left, int right) {
      int i = left, j = right;
      float tmp;
      float pivot = arr[(left + right) / 2];
 
      /* partition */
      while (i <= j) {
            while (arr[i] < pivot)
                  i++;
            while (arr[j] > pivot)
                  j--;
            if (i <= j) {
                  tmp = arr[i];
                  arr[i] = arr[j];
                  arr[j] = tmp;
                  i++;
                  j--;
            }
      };
 
      /* recursion */
      if (left < j)
            quickSort(arr, left, j);
      if (i < right)
            quickSort(arr, i, right);
}
