#ifndef __PEDSEQ_H__
#define __PEDSEQ_H__

#include "PedMarkerData.h"
#include "FamilyLikelihoodLG.h"
#include <iostream>
#include <string>

#include "Argument.h"
#include "IO.h"
#include "tabix.h"
#include "PeopleSet.h"
#include "RangeList.h"
#include "Utils.h"
#include "VCFUtil.h"
#include <cassert>
#include <string>
#include <set>
#include <map>
#include <vector>
#include <algorithm>
#include <iostream>


class PedSeq
{
 public:
  double prior;
  double prior_indel;
  double snv_prior;

  bool isIndel;
  bool isBiallelic;

  int nFounders;
  double theta;
  double theta_indel;
  double tstv_ratio;
  double ts;
  double tv;
  std::string famID;
  int famIdx;

  PedMarkerData *pm;
  PedMarkerData pmd;
  Inheritance_Matrix *im;
  FamilyLikelihoodLG *famlk;
  VarInfo single_var;
  SeqMarkerInfo **single_marker;
 
  std::string out_chr;
  int out_start, out_end;

  std::string cmd;
  std::string chr_prefix;
  std::string vcfInFile;
  std::string vcfOutFile;
  FILE *vcf_fh;
  bool output_all;
  bool disable_fft;
  bool phase_best;
  bool phase_marginal;
  bool single_marginal;
  double cumulative_marginals;
  int PL_idx;
  int GL_idx;
  int DP_idx;

  std::vector<std::string> geno_label;
  char dp_str[100];
  char pl_str[100];

  std::pair<int, int> fam_pid_index_pair;
  bool onfly, trio_only, mixed_vcf_records;
  time_t tstart, tvcf;
  std::string pstart;
  std::string vstart;

  double minDQ;
  std::vector<double> max_bf; 

 public:
  PedSeq();
  ~PedSeq();

  double CalculatePosteriorFromLLK(std::vector<double> &);
  double CalculatePosteriorFromLLK(std::vector<double> &, int post_idx);

  void SetMinDQ(double dq) { minDQ = dq; }
  void SetParam(bool mvr, bool trio) { mixed_vcf_records = mvr; trio_only = trio; }
  void SetChrPrefix(std::string& s) { chr_prefix = s; }
  void SetOnflyFlag(bool flag) { onfly = flag; }
  void SetCmd(std::string &c) { cmd = c; }
  void SetFamilyID(std::string id) { famID = id; }
  void SetFamilyIdx(int idx) { famIdx = idx; }
  void SetIndel(bool flag) { isIndel = flag;}
  bool isTransition(std::string ref, std::string alt);
  void SetTsTvRatio(double tstv){tstv_ratio = tstv; tv = 0.5/(tstv+1); ts=1-2.0*tv;};
  void SetTheta(double t){theta = t; }
  void SetTheta_Indel(double t){ theta_indel = t; }

  void SetBiallelic(bool flag) { isBiallelic = flag; }
  void SetPolyPrior();
  void SetSNVPrior(std::string &ref, std::string &alt);
  void SetInputVCFFile(std::string in) { vcfInFile = in; }
  void SetOutputVCFFile(std::string out){ vcfOutFile = out; }
  void SetPedMarkerData(PedMarkerData *);

  void SetInheritanceMatrix();
  void SetFFTFlag(bool f) { disable_fft = f; }
  void SetPhaseFlag(bool m, bool b) { phase_marginal=m; phase_best=b; }
  void SetSingleMarginal(bool sm) { single_marginal = sm; }
  void CalculateIVMarginals();
  void CalculateDenovoAndOutput2VCF();
  void SetOutputAllFlag(bool flag) { output_all = flag; }
  void SetOutputRange(std::string c, int s, int e) { out_chr = c; out_start = s; out_end = e; }
  void SetCumulativeMarginals(double c){ cumulative_marginals = c; }
  void OutputVCF(FamilyLikelihoodLG &, VarInfo &, FILE*, int famidx, int max_idx);
  void OutputVCF_phased(FamilyLikelihoodLG &, VarInfo &, FILE*, int famidx, int max_idx);
  void OutputVCF_denovo(FILE**, VarInfo &, std::vector<int>&);
  void UpdateMapCounters(int *lpos, int *rpos, int *counter, int *lidx, int *ridx, int single_var_pos);
};

#endif
